#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from __future__ import division
from __future__ import print_function

import json as jasm
from cProfile import Profile
from functools import wraps
from inspect import getfile
from time import time


def cash_it(funk):
    """args-2-return value cache.
    
    This function is particularly useful for when you want that lru-cache,
    but you/one is working with python two.

    :param funk: function to be cached
    :type funk: function
    :returns: wrapped function
    :rtype: function

    """
    cash_money = {}
    @wraps(funk)
    def cash_wrap(*argz):
        """

        """
        if argz in cash_money:
            return cash_money[argz]
        else:
            rv = funk(*argz)
            cash_money[argz] = rv
            return rv
    return cash_wrap

class Jasm(object):
    """Jasm the Grundle Bug"""

    def __init__(self, path):
        self.file_path = path
    def __call__(self, funk):
        """Json saving and loading"""
        fp = self.file_path
        def savings_n_loads(*args, **kwargs):
            """Jasm funk (w)rapper

            """
            if len(args) == 0:
                save_key = "None"
            else:
                save_key = str((args, kwargs.items()))
            try:
                with open(fp) as f:
                    dat_data = jasm.load(f)
            except IOError:
                dat_data = {}
            except ValueError:
                dat_data = {}

            if save_key not in dat_data:
                ret_val = funk(*args, **kwargs)
                dat_data[save_key] = ret_val
                with open(fp, "wb") as f:
                    jasm.dump(dat_data, f, encoding="utf8", sort_keys=True)
            return dat_data[save_key]
        return savings_n_loads
    @staticmethod
    def read(fpath):
        """Jasm load static method

        :param fpath: filepath
        :type fpath: str
        :returns: object/data stored in the json file
        :rtype: object

        """
        with open(fpath) as f:
            return jasm.load(f)
    @staticmethod
    def write(fpath, obj):
        """Jasm dump static method

        :param fpath: filepath
        :type fpath: str
        :param obj: data/object to be saved
        :type obj: object

        """

def cprof(funk):
    """"cProfiling decorator
    
    src: https://zapier.com/engineering/profiling-python-boss/

    :param funk: 

    """
    @wraps(funk)
    def profiled_funk(*args, **kwargs):
        """

        """
        profile = Profile()
        try:
            profile.enable()
            ret_val = funk(*args, **kwargs)
            profile.disable()
        finally:
            print("__CPROFILE__")
            profile.print_stats()
        return ret_val
    return profiled_funk

class tictoc(object):
    """Timing decorator object

    :param runs: # of runs to time over (defaults to 1)

    """

    def __init__(self, runs=1):
        self.runs = runs
    def __str__(self, t_total, funk, args_string):
        str_list = [
            "__TICTOC__",
            "    file: {}".format(getfile(funk)),
            "    funk: {}".format(funk.__name__),
            "    args: {}".format(args_string),
            "    time: {}".format(tictoc.ftime(t_total)),
            "    runs: {}".format(self.runs),
        ]
        return "\n".join(str_list)
    def __call__(self, time_funk, printing=True):
        @wraps(time_funk)
        def time_wrapper(*args, **kwargs):
            """

            """
            self.args = str(args)
            ts = time()
            for i in range(self.runs):
                result = time_funk(*args, **kwargs)
            te = time()
            t_total = (te - ts) / self.runs
            if printing:
                print(self.__str__(t_total, time_funk, str(args)))
            return result
        return time_wrapper
    @staticmethod
    def ftime(t1, t2=None):
        """Formats time string
        
        Formats t1 if t2 is None as a string; Calculates the time and formats
        the time t2-t1 if t2 is not None.

        :param t1: time 1
        :type t1: double
        :param t2: time 2 (Default value = None)
        :type t2: None or double
        :returns: -> formated time string
        :rtype: str

        """
        if t2 is not None:
            return tictoc.ftime((t2 - t1))
        elif t1 == 0.0:
            return "~0.0~"
        elif t1 >= 1:
            return "%.3f s" % t1
        elif 1 > t1 >= 0.001:
            return "%.3f ms" % ((10 ** 3) * t1)
        elif 0.001 > t1 >= 0.000001:
            return "%.3f μs" % ((10 ** 6) * t1)
        elif 0.000001 > t1 >= 0.000000001:
            return "%.3f ns" % ((10 ** 9) * t1)
        else:
            return tictoc.ftime((t2 - t1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from __future__ import division
from __future__ import print_function
from __future__ import with_statement

from pupy.savings_n_loads import load_jasm
from pupy.savings_n_loads import loads
from pupy.savings_n_loads import save_jasm
from pupy.savings_n_loads import savings

try:
    range
except NameError:
    range = xrange

ALL = (
    "degrees_2_radians",
    "disjoint",
    "divisors_gen",
    "expo",
    "fib_r",
    "gcd_it",
    "gcd_r",
    "get_pythag_triple",
    "n_choose_r",
    "partitions_gen",
    "power_mod",
    "pytriple_gen",
    "radians_2_degrees",
    "repermutations",
    "reverse",
    "rfactorial",
    "Trigon",
    "Vuple",
    "loads",
    "load_jasm",
    "safe_path",
    "save_jasm",
    "savings",
    "is_palindrome",
    "string_score",
    "binary_string",
    "OctopusPrime",
    "prime_gen",
    "prime_factorization_gen",
    "prime_factors_gen",
    "is_prime",
    "tictoc",
    "cprof",
    "cash_it",
    "Jasm",
    "chunks",
    "is_permutation",
    "rotate",
    "rotations_gen",
    "digits_list",
    "int_from_digits",
    "iter_product",
)
__all__ = tuple(sorted(ALL))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from __future__ import print_function
from __future__ import with_statement

from codecs import getwriter
from datetime import datetime
from io import open
from itertools import count
from sys import stderr
from time import sleep

try:
    from ujson import dump
    from ujson import load
except:
    from json import dump
    from json import load
from os import path
from os import remove
from msgpack import pack
from msgpack import unpack

def timestamp():
    """Time stamp string w/ format yyyy-mm-ddTHH-MM-SS

    :return: timestamp string
    """
    """Time stamp string w/ format yyyy-mm-ddTHH-MM-SS"""
    return datetime.now().strftime("%Y-%m-%dT%H-%M-%S")

def safe_path(filepath):
    """

    :param filepath: 

    """
    if path.exists(filepath):
        f_bn, f_ext = path.splitext(filepath)
        for n in count(1):
            safe_save_path = f_bn + "_({}).".format(str(n)) + f_ext
            if not path.exists(safe_save_path):
                return safe_save_path
    return filepath

def ensure_save(filepath, n=0):
    try:
        assert path.exists(filepath)
        return True
    except AssertionError:
        sleep(1)
        if n > 5:
            return False
        return ensure_save(filepath, n + 1)

def savings(filepath, string, clobber=True):
    """Save s(tring) to filepath as txt file

    :param filepath: param string:
    :param clobber: return: (Default value = False)
    :param string: 

    """
    if not clobber and path.exists(filepath):
        filepath = safe_path(filepath)
    elif path.exists(filepath):
        remove(filepath)
    try:
        with open(safe_path(filepath) if clobber else filepath, "wb") as file:
            file.write(string.encode("utf-8"))
    except Exception as e:
        raise e

def loads(filepath):
    """Load a (txt) file as a string

    :param filepath: return:

    """
    try:
        with open(filepath, "r", encoding="utf-8") as f:
            return f.read()
    except UnicodeDecodeError as e:
        with open(filepath, "r", encoding="latin2") as f:
            return f.read()

def sjson(filepath, data, min=False):
    """Save json-serial-ize-able data to a specific filepath.

    :param filepath: save filepath
    :param data: json ready data
    :param min: minified format flag
    :return: None
    """
    if type(data) == dict and any(type(val) == bytes for val in data.values()):
        data = {k: str(v, encoding="utf-8") for k, v in data.items()}
    with open(filepath, "wb") as jsonfile:
        if min:
            dump(data, getwriter("utf-8")(jsonfile), ensure_ascii=False)
        else:
            dump(
                data,
                getwriter("utf-8")(jsonfile),
                indent=4,
                sort_keys=True,
                ensure_ascii=False,
                )

def save_jasm(filepath, data, min=False):
    """Alias for sjson"""
    return sjson(filepath, data, min)

def ljson(filepath):
    """Load a json file given a filepath and return the file-data

    :param filepath: path to the jasm file you want to load
    :return: Loaded file contents
    """

    with open(filepath) as infile:
        return load(infile)

def load_jasm(filepath):
    """Alias for ljson"""
    return ljson(filepath)

def spak(filepath, data):
    with open(filepath, 'wb') as pakfile:
        pack(data, pakfile)

def lpak(filepath):
    with open(filepath, 'rb') as pakfile:
        return unpack(pakfile, raw=False)

if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

from os import getcwd
from os import path
from os import sep
from os import walk


def files_gen(dirpath: str = getcwd(), abs=True):
    """Yields paths beneath dirpath param; dirpath defaults to os.getcwd()

    :param dirpath: Directory path to walking down/through.
    :param abs: Yield the absolute path
    :return:
    """
    return (
        fpath if abs else fpath.replace(dirpath, "").strip(sep)
        for fpath in (
            path.join(pwd, file) for pwd, dirs, files in walk(dirpath) for file in files
        )
    )

def dirs_gen(dirpath: str = getcwd(), abs=True):
    """Yields paths beneath dirpath param; dirpath defaults to os.getcwd()

    :param dirpath: Directory path to walking down/through.
    :param abs: Yield the absolute path
    :return:

    """
    return (
        fpath if abs else fpath.replace(dirpath, "").strip(sep)
        for fpath in (pwd for pwd, dirs, files in walk(dirpath))
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from __future__ import division
from __future__ import print_function

from collections import Counter
from collections import deque
from functools import reduce
from operator import mul


def chunks(list, chunk_size):
    """Yields chunks of something slicable with length <= chunk_size

    :param list: list to be broken up
    :type list: list or tuple
    :param chunk_size: size of the chunks
    :type chunk_size: int

    .. docstring::python

        >>> list(chunks([1, 2, 3, 4, 5, 6], 3))
        [[1, 2, 3], [4, 5, 6]]
        >>> list(chunks([1, 2, 3, 4, 5, 6], 2))
        [[1, 2], [3, 4], [5, 6]]
        >>> list(chunks('abcdefghijklmnopqrstuvwxyz', 2))
        ['ab', 'cd', 'ef', 'gh', 'ij', 'kl', 'mn', 'op', 'qr', 'st', 'uv', 'wx', 'yz']
        >>> list(chunks('abcdefghijklmnopqrstuvwxyz', 13))
        ['abcdefghijklm', 'nopqrstuvwxyz']

    """
    return (list[i : i + chunk_size] for i in range(0, len(list), chunk_size))

def is_permutation(a, b):
    """Checks if two integers or lists are permutations lists are permutations

    :param a: possible perumtation of b
    :type a: int or list
    :param b: possible perumtation of a
    :type b: int or list
    :returns: True if a and b are permutations of one another; False otherwise
    :rtype: bool

    """

    if type(a) == int:
        a = digits_list(a)
    if type(b) == int:
        b = digits_list(b)
    return len(a) == len(b) and Counter(a) == Counter(b)

def rotate(rlist, rn=1, left_rotate=True):
    """Rotate a list (rlist) by rn indices to the left or right

    :param rlist: list/toople or slicable to be rotated
    :type rlist: list or tuple
    :param rn: steps bywhich to rotate (Default value = 1)
    :type rn: int
    :param left_rotate: True (default) left rotates; False right rotates.
    :type left_rotate: bool
    :returns: rotated list
    :rtype: list

    .. doctest::python

        >>> rotate([1, 2, 3, 4], left_rotate=True)
        [2, 3, 4, 1]
        >>> rotate([1, 2, 3, 4], left_rotate=False)
        [4, 1, 2, 3]
        >>> rotate([1, 2, 3, 4], rn=4, left_rotate=False)
        [1, 2, 3, 4]

    """
    def _left_rotate(l, n=1):
        """

        :param l: 
        :param n:  (Default value = 1)

        """
        return l[n:] + l[:n]
    def _right_rotate(l, n=1):
        """

        :param l: 
        :param n:  (Default value = 1)

        """
        return l[-n:] + l[:-n]
    return _left_rotate(rlist, rn) if left_rotate else _right_rotate(rlist, rn)

def rotations_gen(rlist):
    """Yields all rotations of a list

    :param rlist:

    .. doctest::python

        >>> for rot in rotations_gen((1, 2, 3, 4)):
        ...     print(rot)
        ...
        (1, 2, 3, 4)
        (4, 1, 2, 3)
        (3, 4, 1, 2)
        (2, 3, 4, 1)

    """
    return ((rlist[-i:] + rlist[:-i]) for i in range(len(rlist)))

def digits_list(number):
    """Returns a list of the digits in num

    :param number: number w/ digits to be listsed
    :type number: int
    :returns: -> digits in a list
    :rtype: list

    .. doctest::python

        >>> digits_list(1111)
        [1, 1, 1, 1]
        >>> digits_list(982)
        [9, 8, 2]
        >>> digits_list(101)
        [1, 0, 1]
        >>> digits_list(123)
        [1, 2, 3]

    """

    digits = deque()
    for _ in range(len(str(number))):
        number, r = divmod(number, 10)
        digits.appendleft(r)
    return list(digits)

def int_from_digits(digits):
    """Converts an iterable of digits digits to a number
    
    The iteratble can be ints or strings/chars

    :rtype: int

    .. doctest::python

        >>> int_from_digits([3, 2, 1])
        321
        >>> int_from_digits([1, 1, 1, 1, 2, 3])
        111123
        >>> int_from_digits([1, 2, 3])
        123

    """
    return sum(digits[len(digits) - i - 1] * 10 ** i for i in range(0, len(digits), 1))

def iter_product(l):
    """Product of all the elements in a list or tuple

    :param l: list with integer elements
    :returns: product of all the elements in a list
    :rtype: int

    .. doctest::python

        >>> iter_product([1, 2, 3, 4])
        24
        >>> iter_product(tuple([1, 2, 3, 4]))
        24
        >>> iter_product([-1, -2, -3, 4])
        -24

    """
    return reduce(mul, l)

if __name__ == "__main__":
    import doctest

    doctest.testmod()
